import pytest

from avrae_ls import argparser
from avrae_ls.argparser import Argument, EphemeralArgument


def test_argparse_arg_respects_ephemeral_flag():
    assert argparser._argparse_arg("d", None, True, 0, parse_ephem=True) == Argument("d", True, 0)
    assert argparser._argparse_arg("d", "1", True, 0, parse_ephem=True) == EphemeralArgument("d", True, 0, 1)
    assert argparser._argparse_arg("d", "1", True, 0, parse_ephem=False) == Argument("d1", True, 0)


def test_argparse_iterator_matches_avrae_dfa():
    args = ["d", "d1", "adv1", "adv", "-i"]
    parsed = list(argparser._argparse_iterator(args, parse_ephem=True))
    assert parsed == [
        Argument("d", True, 0),
        EphemeralArgument("d", True, 1, 1),
        EphemeralArgument("adv", True, 2, 1),
        Argument("adv", True, 3),
        Argument("i", True, 4),
    ]


@pytest.mark.parametrize(
    ("args", "parse_ephem", "expected"),
    [
        (["12345", "this is junk", "!*&^#&(*#"], True, []),
        (["-d"], True, [Argument("d", True, 0)]),
        (["-d1"], True, [EphemeralArgument("d", True, 0, 1)]),
        (["-d", "5"], True, [Argument("d", "5", 0)]),
        (["-d1", "5"], False, [Argument("d1", "5", 0)]),
        (["-d", "-1d6"], True, [Argument("d", "-1d6", 0)]),
        (["-d1", "-1d6"], False, [Argument("d1", "-1d6", 0)]),
        (["-t", "d5"], True, [Argument("t", "d5", 0)]),
        (["-t", "-i"], True, [Argument("t", True, 0), Argument("i", True, 1)]),
        (["-t", "-t"], True, [Argument("t", True, 0), Argument("t", True, 1)]),
    ],
)
def test_argparse_iterator_edge_cases(args, parse_ephem, expected):
    parsed = list(argparser._argparse_iterator(args, parse_ephem=parse_ephem))
    assert parsed == expected



from pathlib import Path

from avrae_ls.api import AliasContextAPI, CharacterAPI, CombatAPI
from avrae_ls.config import AvraeLSConfig


def _profile():
    cfg = AvraeLSConfig.default(Path("."))
    return cfg.profiles[cfg.default_profile]


def test_character_api_mock_schema():
    profile = _profile()
    ch = CharacterAPI(profile.character)
    assert ch.name == "Aelar Wyn"
    assert ch.stats.prof_bonus == 3
    assert ch.levels.total_level == 5
    assert ch.skills.athletics.value >= 5
    assert ch.saves.get("str").value >= 5
    assert ch.coinpurse.total > 0
    assert ch.cc_exists("Hit Dice")
    start = ch.get_cc("Hit Dice")
    ch.mod_cc("Hit Dice", -1)
    assert ch.get_cc("Hit Dice") == start - 1
    ch.reset_hp()
    assert ch.hp == ch.max_hp
    assert ch.spellbook.can_cast("Cure Wounds", 1)


def test_combat_api_mock_schema():
    profile = _profile()
    cmb = CombatAPI(profile.combat)
    assert cmb.round_num >= 1
    assert cmb.me is not None
    assert cmb.get_combatant(cmb.me.name) is not None
    assert cmb.get_group("Goblins") is not None
    assert cmb.get_metadata("scene") == "Forest road"
    cmb.set_metadata("note", "testing")
    assert cmb.get_metadata("note") == "testing"
    cmb.end_round()
    assert cmb.turn_num == 0


def test_ctx_api_mock_schema():
    profile = _profile()
    ctx = AliasContextAPI(profile.ctx)
    assert ctx.guild.name
    assert ctx.channel.name
    assert ctx.author.display_name
    assert len(ctx.author.get_roles()) >= 1


def test_character_attack_list_helpers():
    char = CharacterAPI(
        {
            "attacks": [
                {"name": "Claw", "verb": "slashes", "raw": {"damage": "1d4+2 slashing"}},
                {"name": "Bite"},
            ]
        }
    )
    attacks = char.attacks
    assert len(attacks) == 2
    first = attacks[0]
    assert first.verb == "slashes"
    assert str(first) == "Claw slashes: 1d4+2 slashing"
    assert str(attacks).splitlines()[-1] == "Bite"


def test_character_consumables_normalized_and_mutated():
    char = CharacterAPI({"consumables": [{"name": "Arrows", "value": 20, "max": 40}]})
    assert char.cc_exists("Arrows")
    assert char.get_cc("Arrows") == 20
    assert char.mod_cc("Arrows", -5) == 15
    assert char.set_cc("Arrows", value=10, maximum=50) == 10
    assert char.get_cc_max("Arrows") == 50
    assert char.delete_cc("Arrows") is not None
    assert not char.cc_exists("Arrows")


def test_spellbook_slot_helpers_and_casting():
    book = CharacterAPI(
        {
            "spellbook": {
                "slots": {1: 2, 2: 1},
                "max_slots": {1: 3, 2: 2},
                "max_pact_slots": 2,
                "num_pact_slots": 1,
                "spells": [{"name": "Magic Missile", "prepared": True}],
            }
        }
    ).spellbook

    assert book.slots_str(1) == "2/3"
    assert book.can_cast("Magic Missile", 1)
    assert book.find("magic missile")[0].prepared is True

    book.cast("Magic Missile", 2)
    assert book.get_slots(2) == 0

    book.reset_slots()
    assert book.get_slots(1) == 3

    book.reset_pact_slots()
    assert book.num_pact_slots == 2



from pathlib import Path

from lsprotocol import types
from pygls.workspace import Workspace

from avrae_ls.symbols import (
    build_symbol_table,
    document_symbols,
    find_definition_range,
    find_references,
    range_for_word,
)
from avrae_ls.server import AvraeLanguageServer, on_rename


def test_build_symbol_table_and_document_symbols():
    source = "x = 1\n\ndef foo():\n    return x\n"
    table = build_symbol_table(source)

    assert any(entry.name == "x" for entry in table.entries)
    assert any(entry.name == "foo" for entry in table.entries)

    doc_symbols = document_symbols(source)
    assert {s.name for s in doc_symbols} == {"x", "foo"}


def test_find_definition_range():
    source = "x = 1\ny = x + 2\n"
    table = build_symbol_table(source)
    rng = find_definition_range(table, "x")
    assert rng is not None
    assert rng.start.line == 0
    assert rng.start.character == 0


def test_find_references_includes_definition_and_usages():
    source = "x = 1\ny = x + x\n"
    table = build_symbol_table(source)
    ranges = find_references(table, source, "x", include_declaration=True)
    starts = {(r.start.line, r.start.character) for r in ranges}
    assert (0, 0) in starts  # declaration
    assert (1, 4) in starts
    assert (1, 8) in starts
    assert len(ranges) == 3  # definition deduped


def test_find_references_respects_inline_draconic_offsets():
    source = "text <drac2>x = 1; y = x</drac2>\n"
    table = build_symbol_table(source)
    ranges = find_references(table, source, "x", include_declaration=True)
    starts = {(r.start.line, r.start.character) for r in ranges}
    assert (0, 12) in starts  # first x
    assert (0, 23) in starts  # second x
    assert len(ranges) == 2


def test_find_references_omits_declaration_when_requested():
    source = "x = 1\ny = x + x\n"
    table = build_symbol_table(source)
    ranges = find_references(table, source, "x", include_declaration=False)
    starts = {(r.start.line, r.start.character) for r in ranges}
    assert (0, 0) not in starts
    assert len(ranges) == 2


def test_range_for_word_returns_selection():
    source = "foo = 1\nbar = foo\n"
    pos = types.Position(line=1, character=6)
    rng = range_for_word(source, pos)
    assert rng is not None
    assert rng.start.line == 1
    assert rng.start.character == 6
    assert rng.end.character == 9


def test_on_rename_returns_workspace_edit_for_symbol():
    server = AvraeLanguageServer()
    server.load_workspace(Path("."))
    server.protocol._workspace = Workspace(None, sync_kind=types.TextDocumentSyncKind.Incremental)
    uri = "file:///test.alias"
    source = "x = 1\ny = x\n"
    server.protocol.workspace.put_text_document(
        types.TextDocumentItem(uri=uri, language_id="python", version=1, text=source)
    )

    params = types.RenameParams(
        text_document=types.TextDocumentIdentifier(uri=uri),
        position=types.Position(line=1, character=4),
        new_name="renamed",
    )
    edit = on_rename(server, params)
    assert edit is not None
    edits = edit.changes[uri]
    updated = _apply_edits(source, edits)
    assert "renamed = 1" in updated
    assert "y = renamed" in updated


def _apply_edits(source: str, edits: list[types.TextEdit]) -> str:
    lines = source.splitlines()
    # Apply edits in reverse document order to avoid shifting offsets
    for edit in sorted(edits, key=lambda e: (e.range.start.line, e.range.start.character), reverse=True):
        start = edit.range.start
        end = edit.range.end
        if start.line == end.line:
            line = lines[start.line]
            lines[start.line] = line[: start.character] + edit.new_text + line[end.character :]
        else:
            before = lines[start.line][: start.character]
            after = lines[end.line][end.character :]
            lines[start.line : end.line + 1] = [before + edit.new_text + after]
    return "\n".join(lines) + ("\n" if source.endswith("\n") and not source.endswith("\n\n") else "")



import json
from pathlib import Path

import pytest

from avrae_ls.config import AvraeServiceConfig, CONFIG_FILENAME, load_config


def _write_config(tmp_path: Path, data: dict) -> Path:
    config_path = tmp_path / CONFIG_FILENAME
    config_path.write_text(json.dumps(data))
    return config_path


def test_load_config_env_substitution(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setenv("AVRAE_TOKEN", "shh-its-a-secret")
    monkeypatch.setenv("AVRAE_BASE", "https://env.example.invalid")
    _write_config(
        tmp_path,
        {
            "avraeService": {
                "baseUrl": "${AVRAE_BASE}",
                "token": "$AVRAE_TOKEN",
            },
        },
    )

    cfg, warnings = load_config(tmp_path)

    assert warnings == []
    assert cfg.service.base_url == "https://env.example.invalid"
    assert cfg.service.token == "shh-its-a-secret"


def test_load_config_workspace_root_substitution(tmp_path: Path):
    _write_config(
        tmp_path,
        {
            "varFiles": ["${workspaceRoot}/vars.json"],
        },
    )

    cfg, warnings = load_config(tmp_path)

    assert warnings == []
    assert cfg.var_files == (tmp_path / "vars.json",)


def test_load_config_missing_env_warns(tmp_path: Path):
    _write_config(
        tmp_path,
        {
            "avraeService": {
                "baseUrl": "${MISSING_AVRAE_BASE}",
                "token": "$MISSING_TOKEN",
            },
        },
    )

    cfg, warnings = load_config(tmp_path)

    assert any("MISSING_AVRAE_BASE" in warning for warning in warnings)
    assert any("MISSING_TOKEN" in warning for warning in warnings)
    assert cfg.service.base_url == AvraeServiceConfig.base_url
    assert cfg.service.token is None
